; This code version is for the UPS-14500-4040-HAT hardware version. The major change
; is to the battery cutoff voltage -- decreased from 3.0V to 2.75V.

; PIC16F18313 Configuration Bit Settings

; Assembly source line config statements

#include "p16f18313.inc"

; CONFIG1
; __config 0xDFEC
 __CONFIG _CONFIG1, _FEXTOSC_OFF & _RSTOSC_HFINT1 & _CLKOUTEN_OFF & _CSWEN_OFF & _FCMEN_OFF
; CONFIG2
; __config 0xF713
 __CONFIG _CONFIG2, _MCLRE_ON & _PWRTE_ON & _WDTE_OFF & _LPBOREN_ON & _BOREN_OFF & _BORV_LOW & _PPS1WAY_OFF & _STVREN_OFF & _DEBUG_OFF
; CONFIG3
; __config 0x3
 __CONFIG _CONFIG3, _WRT_OFF & _LVP_ON
; CONFIG4
; __config 0x3
 __CONFIG _CONFIG4, _CP_OFF & _CPD_OFF




 ; General Notes:
 ; Configure internal oscillator for 1MHz (250kHz system clock). No external
 ; osccillator needed. This is set in the configuration bits. 
 ;
 ; This part uses interrupts extensively and has a 16-level stack.
 ;
 ; Use the low-power sleep mode since current drain in sleep is a priority and
 ; wake from sleep does not have to happen quickly.
 ;
 ; These are the unneeded, and therefore disabled, components:
 ; Comparator1, Comparator2 (disabled at reset)
 ; Comparator voltage reference
 ; Timer0-2  - TBD
 ; DAC1
 ; ZCD1
 ; TIM
 ; CCP1, CCP2
 ; CLC1-4
 ; CWG1
 ; EUSART1-2 
 ; PWM3-6
 ; NVM
 ; Watchdog Timer - TBD. WDT can be disabled during sleep so 
 ; try to enable it when things are working smoothly.
 ;
 ; The ADC is 10-bit, but will only need 8-bits. Uses RA5 as input and measures 
 ; internal FVR=4.096V. No external reference required. 
 ;
 ; Weak pullups can be individually assigned. Don't use them - 200uA!
 ; 
 ; Pin Assignments:
 ;  1 -  VDD
 ;  2 -  RA5: ADC input
 ;  3 -  RA4: Digital output - PowerOFF signal
 ;  4 -  MCLR_, Master reset, digital schmidt trigger input, 10k to VDD.
 ;  5 - RA2: I2C SDA.
 ;  6 - RA1: I2C_SCL
 ;  7 - RA0: Digital schmidt trigger input - ACPR,  wake on change
 ;  8 - VSS


;***** Define addresses of the General Purpose Registers ****
; lots of general purpose registers available starting at 0x20 or
; use Common Ram (available to all banks - except bank0) starting at 0x70
COUNT1	equ 0x20    ;First counter for delay loops
COUNT2	equ 0x21    ;Second counter for delay loops
COUNT3	equ 0x22    ;Third counter for delay loops
SCRATCH	equ 0x23    ;Scratchpad variable for everthing else.
ADC_SUM equ 0x25    ; rolling average summation - reserve two bytes
ADC_SUMH equ 0x26
AVERAGE equ 0x27    ; rolling average result LSB
AVERAGEH  equ 0x28
AVG_COPY equ 0x29   ; rolling average copy used for I2C reporting.
I2CSTAT	equ 0x2a    ; status register for I2C interface
		    ; bit0 = PWRGOOD
		    ; bit1 = BATLOW (handshake to Pi)
		    ; bit2 = SHUTDN
smp_buf	equ 0x2b    ; pointer to the start of the 8 rolling average registers
	
RES_VECT  CODE    0x0000            ; processor reset vector

    goto	DISPER
 
; *** Process Interrupts Here *** 

    ORG 0004h		    ; Interrupt service address
    ; service I2C first. The wake from sleep should not generate an interrupt
    ; but it is put here just in case...

	banksel	0    ; 
	btfsc	PIR1,3	    ; test the SSP1IF bit to check for I2C interrupt
	call	I2C	    ; service I2C interface
	
	; the following should not execute if wake on change works
	;movlb	IObank	    ; 
	;btfsc	IOCAF,0	    ; check for interrupt on RA0/ACPR
	;call	ACPRINT	    ; service ACPR interrupt
	
	retfie		    ; end of interrupt service routine

;**** Configure Clocks ***
; No changes yet. Internal osc runs @1MHz so system clock is 250KHz. It is 
; set this high to handle I2C timing at 100KHz.
	
;**** Disable unused peripherals ****
DISPER	
	banksel	INTCON	    
	clrf	INTCON	    ; suspend interrupts
	banksel	PMD0	
	movlw	b'00000110' ; enable Peripheral system clock network, enable
	movwf	PMD0	    ; fixed voltage ref, disable NVM module, disable
			    ; clock ref, enable interrupt on change module.
	movlw	b'10000111' ; disable NCO, Disable TMR2-0
	movwf	PMD1
	movlw	b'01000110' ; disable DAC, enable ADC, disable comparators
	movwf	PMD2	    ; 
	movlw	b'01110011' ; disable all PWMs and CCPs
	movwf	PMD3
	movlw	b'00100000' ; disable USART1-2, enable MSSP
	movwf	PMD4
	movlw	0x07	    ; disable all CLC modules and delta-sigma modulator
	movwf	PMD5
	
;**** Configure Interrupts ***
	
	banksel	IOCAP	    ; (changed Hat version from neg to pos WOC.)
	clrf	IOCAN	    ; disable Interrupt on change neg edge for PORTA
			    ; enable RA0/ACPR Interrupt-on-change 
	bsf	IOCAP,0	    ; for positive edge only
	;clrf	IOCCP	    ; disable IOC pos edge for PORTC (not in 18313)
	;clrf	IOCCN	    ; disable IOC neg edge for PORTC (not in 18313)
	; PIE1-PIE7 default to disable peripheral modules
	banksel	PIE0    
	clrf	PIE0	    ; disable Interrupt-on-change until sleep
	bsf	PIE1,3	    ; enable I2C interrupt
	
;*** Setup Sleep Mode and DOZE operation ***
	banksel	VREGCON    
	movlw	0x03	    ; Low power sleep mode enabled
	movwf	VREGCON	    ; 
	banksel	CPUDOZE	     
	clrf	CPUDOZE	    ; disable DOZE mode until needed later
	
;*** Assign I2C SCL and SDA to pins ***
	
	banksel	SSP1CLKPPS	     
	movlw	0x01	    ; assign RA1 as 
	movwf	SSP1CLKPPS  ; I2C_SCL input
	movlw	0x02	    ; assign RA2 as 
	movwf	SSP1DATPPS  ; I2C_SDA input
	banksel	RA1PPS	    
	movlw	b'11000'    ; assign RA1 as 
	movwf	RA1PPS	    ; output for I2C_SCL
	movlw	b'11001'    ; assign RA2 
	movwf	RA2PPS	    ; I2C_SDA output

;*** Setup ADC ***
	
	banksel	FVRCON	    ; 
	movlw	b'01000011' ; FVR disabled temporarily, Temp indicator disabled,
	movwf	FVRCON	    ; comparator FVR is off, ADC FVR = 4.096V
	banksel	ADCON0	    
	movlw	b'00110011' ; Format:left Justified, Clock:RC Osc, VREF=FVR
	movwf	ADCON1
	movlw	b'00010100'  ; Input=RA5 (ANA5), ADC Off
	movwf	ADCON0	    ;
	
;*** Configure I2C interface (MSSP module) ***
; Slave mode. Only NACKing address for now.
BEGIN
	banksel	0	    ; set to bank0
	clrf	INTCON	    ; suspend interrupts temporarily
	banksel	SSP1STAT
	movlw	b'10000000' ; slew rate control disabled, smbus levels?
	movwf	SSP1STAT    ; SSP1 Status reg.
	movlw	b'00110110' ; Enable I2C mode/pins, CKP=1, set I2C as 
	movwf	SSP1CON1    ; slave using 7-bit address
	movlw	b'00000000' ; in slave mode: general call interrupt disabled,
	movwf	SSP1CON2    ; ACKDT=0, ACKEN=0, RCEN=0, PEN=0, RSEN=0, SEN=0
	movlw	b'00011000' ; ACKtie=0, PCIE=0, SCIE=0, BUF overwrite OK,
	movwf	SSP1CON3    ; SDA hold time 300ns, Collision disabled, 
			    ; AHEN disabled, DHEN= disabled.
	movlw	0x6C	    ; I2C Address should be 0x36
	movwf	SSP1ADD	    ;
	clrf	SSP1BUF	    ;
	banksel	0	    ; set to bank0
	clrf	I2CSTAT	    ; clear I2C status register.

;**** Initialize PORTS ****	
			    ; NOTE: RA0/ACPR is set here to STrigger but later
			    ; it will be reprogrammed to TTL to handle lower
			    ; voltage levels.
	;banksel SLRCONA    
	;clrf	SLRCONA	    ; use slew rate control which is default at POR.
	;clrf	SLRCONC	    ; not in 18313
	banksel	ANSELA
	movlw	b'00100000'
	movwf	ANSELA	    ; set RA5 to analog input for ADC.
	;clrf	ANSELC	    ; there are no analog inputs on PORTC (not in 18313)
	banksel	WPUA
	clrf	WPUA	    ; No weak pullups. 
	;clrf	WPUC	    ; no weak pullups on PORTC (no WPUC on 18313)
	banksel	INLVLA
	movlw	b'00101001'
	movwf	INLVLA	    ; MCLR_ and RA0/ACPR are Schmidt trigger levels.
	;clrf	INLVLC	    ; all inputs are TTL levels (not in 18313) 
	; RA0 & RA3  must be input. Clock pins set by configuration bits.
	; setup I/O first -- keep outputs Low!
	banksel	LATA	    ; set to bank 0
	movlw	b'111111'   ; Initialize PortA. Set PWROFF = 1 to disable
	movwf	LATA	    ; the LTC4040 until AC is present.
	;clrf	PORTC	    ; Init PortC (not in 18313)
	banksel	TRISA
	movlw	b'00101111'   ; load TRISA, set PORTA inputs: RA0/ACPR as ST 
	movwf	TRISA	    ;  and RA1/SCL RA2/SDA as I2C, RA5 as ADC input
	;clrf	TRISC	    ; PORTC has no inputs. (not in 18313)

; **** Initialize rolling average of battery voltage
	banksel	0
	clrf	ADC_SUM	    ; initialize ADC_SUM to zero
	clrf	ADC_SUMH    ; ditto
	movlw	smp_buf	    ; initialize pointer to rolling average block 
	movwf	FSR0
CBUF	clrf	INDF0	    ; clear the buffer registers
	incf	FSR0,1
	movlw	0x33	    ; smp_buf + 8 = end of buffer
	xorwf	FSR0,0
	btfss	STATUS,2    ; end of buffer test
	goto	CBUF
	movlw	smp_buf	    ; set pointer to rolling average block 
	movwf	FSR0
	
	movlw	0x08	    ; put counter into SCRATCH register
	movwf	SCRATCH
Filbuf	call	BatCheck    ; measure battery 8 times to 
	decfsz	SCRATCH,1    ; fill up the rolling average block
	goto	Filbuf

	
;**** Enable interrupts ****
	banksel SSP1CON1    
	bsf	SSP1CON1,4  ; set the CKP bit to release SCL stretch
	banksel	0	    ; set to bank0
	movlw	b'11000000' ; enable peripheral interrupts for I2C and
	movwf	INTCON	    ; enable general interrupts
	call	Delay1	    ; wait 1 second for things to settle out.
	
	
;**** Check ACPR ****
	
Start	
	banksel	PORTA	    
	btfsc	PORTA,0	    ; test state of RA0/ACPR (changed for Hat design)
	goto	AplPwr	    ; AC is present
			    
;**** ACPR is Low -- AC is absent -- goto SLEEP ****
	
Sleepy	
	banksel	INLVLA
	bcf	INLVLA,0	    ; change RA0/ACPR to ST level before sleep
	banksel	0	    ; set to bank0
	clrf	INTCON	    ; disable general and peripheral interrupts
			    ; this still allows for wake-on-change
	banksel	IOCAF	    ; 
	clrf	IOCAF	    ; clear IOC flag for RA0/ACPR (just in case)
	banksel	PMD4
	bsf	PMD4,1	    ;  disable MSSP
	banksel	PIE0    
	movlw	0x10	    ; enable Interrupt-on-change
	movwf	PIE0
	sleep
	NOP
	; wake-on-change should not generate an interrupt so the following
	; should clear the flag and disable the interrupt on change for PORTA
	banksel	PIE0    
	clrf	PIE0	    ; disable Interrupt-on-change
	banksel	IOCAF	    
	clrf	IOCAF	    ; clear IOC flag for RA0/ACPR
	banksel	PMD4
	bcf	PMD4,1	    ;  enable MSSP
	goto	BEGIN	    ; go back to beginning. Global interrupts will be
			    ; disabled temporarily after jump to BEGIN.
			    ; NOTE!!! After wake from sleep the MSSP module 
			    ; must be re-initialized or it doesn't seem to work.
	 
;**** ...apply power ****
AplPwr	bcf	PORTA,4	    ; Clear RA4/PWROFF to enable the LTC4040
	banksel	INLVLA
	bcf	INLVLA,0	    ; change RA0/ACPR to TTL level

W81s	;banksel	PORTC
	;bsf	PORTC,5	    ; set RC5/PWRFLAG (not in 18313)
	banksel	0
	bsf	I2CSTAT,0   ; set the PWRGOOD bit in the I2C status register
	btfsc	I2CSTAT,2   ; test if Pi has set SHUTDN bit
	goto	Piconf	    ; the Pi says shut 'er down
	banksel	CPUDOZE
	movlw	b'01110111' ; IDLEN=0, DOZE enabled, ROI enabled
	movwf	CPUDOZE	    ; DOZE on exit enabled, DOZE ratio = 1:256
	call	DelayDOZE   ; Wait 1 second with slower clock to reduce current
	banksel	CPUDOZE
	clrf	CPUDOZE	    ; disable DOZE modes
	call	BatCheck    ; measure battery to make available to I2C

;**** Is the power bad? ***
PBADCHK	btfsc	PORTA,0	    ; test state of RA0/ACPR
	goto	W81s	    ; PWR is good. Wait 1 sec and test again.
	;bcf	PORTC,5	    ; PWR is bad, clear RC5/PWRFLAG (not in 18313)
	bcf	I2CSTAT,0   ; clear the PWRGOOD bit
	
;**** Test for SHUTDN signal from Pi ****
SDtst	btfsc	I2CSTAT,2   ; test if Pi issued SHUTDN
	goto	Piconf	    ; the Pi says shut 'er down

;**** Check battery voltage to see if shutdown necessary ***
	call	BatCheck    ; Test battery voltage
	btfss	I2CSTAT,1   ; battery is low. Shutdown power.
	goto	PBADCHK	    ; battery is good, recheck ACPR

;**** Commited to shutdown now ***
	; wait 3-5 seconds to let Pi know that shutdown is imminent. This assumes
	; that the Pi will be reading the I2C info more rapidly as the battery
	; voltage gets lower. If the SHUTDN bit is set then just go to shutdown
Shtdwn	movlw	d'5'	    ; put 5 into SCRATCH
	movwf	SCRATCH
W85	btfsc	I2CSTAT,2   ; test if Pi issued SHUTDN
	goto	Piconf	    ; Pi confirms
	call	Delay1
	decfsz	SCRATCH	    ; wait 5 seconds for Pi to confirm, then shutdown
	goto	W85
Piconf	bsf	I2CSTAT,1   ; assert BATLOW as confirm flag to Pi
	call	Delay20	    ; wait 20 sec for Pi to shutdown
	bsf	PORTA,4	    ; turn off RA4/PWROFF (the booster)
	bcf	I2CSTAT,1   ; turn off BATLOW bit
	call	Delay5	    ; wait 5 sec for the Pi to power off.

;****Now go back to the start of the program
	banksel	INLVLA
	bsf	INLVLA,0    ; change RA0/ACPR to ST level
	goto	Start       ; loop forever
	
;**** Start of subroutine section **** 

BatCheck
;**** perform ADC conversion and test if battery high or low ****
	banksel	FVRCON		
	bsf	FVRCON,7	; FVR enabled
FVRGO	btfss	FVRCON,6	; wait for FVRRDY bit
	goto	FVRGO
	banksel	ADCON0		
	bsf	ADCON0,0	; turn on ADC
	call	SampleTime	; Acquisition delay
	banksel	ADCON0		
	bsf	ADCON0,1	; start ADC conversion
W84ADC	btfsc	ADCON0,1	; wait for end of conversion
	goto	W84ADC
	bcf	ADCON0,0	; turn off ADC
	banksel	FVRCON		; 
	bcf	FVRCON,7	; turn off FVR
	call	rollAvg		; add the result to the rolling average
	banksel	0		; set the bank back to 0
	movlw	0x54		; add 0x54 (84 decimal) to the AVG_COPY and
	addwf	AVG_COPY,0	; store result in W reg
	btfss	STATUS,0	; overflow = yes or no?
	goto	BatLo		; No? - Battery is below low limit
	bcf	I2CSTAT,1	; Yes? - BATLOW = 0
	return
BatLo	bsf	I2CSTAT,1	; set BATLOW = 1
	return	
 
;**** Perform rolling average on ADC data ****
rollAvg	
	banksel	0	    ; 
	movf	INDF0,0	    ; subtract LSB of old sample from running total
	subwf	ADC_SUM,1   ;
	btfss	STATUS,0    ; check for borrow from subtraction
	decf	ADC_SUMH,1  ; subtract borrow from running total MSB  
	banksel	ADRESH
	movf	ADRESH,0    ; put the ADC result in W
	banksel	0
	movwf	INDF0	    ; put the ADC result into the rolling avg buffer
	addwf	ADC_SUM,1   ; add ADC result to running total
	btfsc	STATUS,0    ; test for carry from addition
	incf	ADC_SUMH,1    ; add the carry to the MSB of the running total 
	incf	FSR0,1	    ; advance buffer pointer to next sample
	movlw	0x33	    ; smp_buf + 8 = end of buffer
	xorwf	FSR0,0	    ; if end of buffer is reached
	movlw	smp_buf
	btfsc	STATUS,2    
	movwf	FSR0	    ; reset FSR to start of buffer
	; now divide running total by 8
	movf	ADC_SUM,0   ;put LSB of running total in W
	movwf	AVERAGE	    ; move W in AVERAGE LSB
	movf	ADC_SUMH,0 ; put MSB of running total in W
	movwf	AVERAGEH   ; move W to AVERAGE MSB
	bcf	STATUS,0    ; clear carry bit
	rrf	AVERAGEH,1 ; shift right three times:
	rrf	AVERAGE,1
	bcf	STATUS,0    ; clear carry bit
	rrf	AVERAGEH,1
	rrf	AVERAGE,1
	bcf	STATUS,0    ; clear carry bit
	rrf	AVERAGEH,1
	rrf	AVERAGE,1
	movf	AVERAGE,0   ; put average result in W
	movwf	AVG_COPY    ; move average result to AVG_COPY
	return
	
SampleTime
	banksel	0	    ; set to bank0
	movlw	d'2'	    ; X
	movwf	COUNT1	    ; put X into COUNT1
LoopST	decfsz	COUNT1,1    ; counts down from X
	goto	LoopST
	return		    ; total delay = X + 7 * 4 usec
  
DelayDOZE		    ; total count = 250k/256 = 980 = 1 second
	banksel	0	    ; set to bank0
			    ; (COUNT1 x 3 + 5)COUNT2  = 979
	movlw	d'11'	    ; put 11 
	movwf	COUNT2	    ; into COUNT2
Loop2D	movlw	d'28'	    ; put 28
	movwf	COUNT1	    ; into COUNT1	    		    
Loop1D	decfsz	COUNT1,1    ; counts down from 28
        goto	Loop1D       ;
        decfsz	COUNT2,1    ; counts down from 11
	goto	Loop2D	    ;
	 
	return
	
Delay1			    ; total count = 250k = 1 second
	banksel	0	    ; set to bank0
	movlw	d'2'	    ; put 2 into COUNT3
	movwf	COUNT3	    ; COUNT3 x (COUNT1 x 3 + 3)(COUNT2 + 5) = 250368
	movlw	d'255'	    ; put 255 
	movwf	COUNT1	    ; into COUNT1
Loop2	movlw	d'158'	    ; put 158
	movwf	COUNT2	    ; into COUNT2	    
	
Loop1	decfsz	COUNT1,1    ; counts down from 255
        goto	Loop1       ;
        decfsz	COUNT2,1    ; counts down from 158
	goto	Loop1	    ;
	decfsz	COUNT3,1    ; counts down from 2
	goto	Loop2	    ; 
	return

Delay5
	banksel	0	    ; set to bank0
	movlw	0x05	    ; 5 x 1 = 5 seconds
	movwf	SCRATCH	    ; into SCRATCH register
Loop5	
	call	Delay1	    ; 1 sec delay
	decfsz	SCRATCH,1   ;
        goto	Loop5       ;
	return

Delay20
	banksel	0	    ; set to bank0
	movlw	d'20'	    ; 20 x 1 = 20 seconds
	movwf	SCRATCH	    ; into SCRATCH reg.
Loop20	
	call	Delay1	    ; 1 sec delay
	decfsz	SCRATCH,1   ;
        goto	Loop20       ;
	return

;*** Service ACPR interrupt ***
ACPRINT
	banksel	PIE0    
	clrf	PIE0	    ; disable interrupt-on-change for RA0/ACPR
	banksel	IOCAF	    
	clrf	IOCAF	    ; clear IOC flag for RA0/ACPR
	return
	
;*** Service I2C interface ***
; Access to the R/W I2CSTAT register and read the ADC_COPY register
; General protocol is the smbus read_word_data(), which uses this pattern:
;   Address_W->CommandRegister->Address_R->Byte1->Byte2
;
; Note that the CommandRegister is can only be written with two values: 0x00 or 0x04
; and that these bytes map only to the SHUTDN bit of I2CSTAT register.
; Byte1 is simply reading the I2CSTAT register and byte2 is a copy of the last 
; ADC conversion result.
	;   Byte1 = status. PWRGOOD(0), BATLOW(1), SHUTDOWN(2) confirmation
	;   Byte2 = ADC data (ADC_COPY) 8-bits is good enough
; The interface should also allow a single smbus byte write: write_byte_data()
;   Adress_W->CommandRegister.
	; Write Protocol:
	; 1 Byte: shutdown bit. Shutdown active after next I2C read.
I2C
	banksel	PIR1	    
	bcf	PIR1,3	    ; clear the SSP1IF interrupt flag
	banksel	SSP1STAT    
	btfss	SSP1STAT,5  ; test for address or data
	goto	ADDR	    ; it's address
	btfss	SSP1STAT,2  ; It's data, get CMD if Write, send byte2 if Read
	goto	WRTBYTE	    ; get the CMD byte 
	banksel	0	    ; process second byte read:
	movf	AVG_COPY,0  ; put ADC data into W reg then
	banksel	SSP1BUF	    ; load I2C buffer with second byte...
	movwf	SSP1BUF	    ; 
	goto	DONE	    ; ...and transmit it.

WRTBYTE	btfss	SSP1BUF,2   ; test for SHUTDN bit
	goto	CLRSD
	banksel	0
	;btfss	I2CSTAT,0   ; ignore shutdown command if PWRGOOD = 1
	bsf	I2CSTAT,2   ; Set the SHUTDN bit in I2CSTAT register
	goto	DEND
CLRSD	
	banksel	0
	bcf	I2CSTAT,2   ; Clear the SHUTDN bit in I2CSTAT register
DEND	
	banksel	SSP1BUF	    
	clrf	SSP1BUF	    ; clear the buffer
	goto	DONE
	
ADDR	
	movf	SSP1BUF,0   ; read the buffer to clear BF bit
	btfss	SSP1STAT,2  ; Test for Read/Write_
	goto	DONE	    ; if Write then exit and wait for next byte
	banksel	0	    ; if Read, then put byte1 in buffer
	movf	I2CSTAT,0   ; put data to send in W register
	banksel	SSP1BUF	    
	movwf	SSP1BUF	    ; load buffer from W register

DONE	bsf	SSP1CON1,4  ; set the CKP bit to release SCL stretch

	return

    END